# -*- coding: utf-8 -*-

"""
    sijax.plugin.comet.CometResponse
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Provides the CometResponse class, an instance of which
    is passed as the first argument to all Comet functions.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from ...response.StreamingIframeResponse import StreamingIframeResponse


class CometResponse(StreamingIframeResponse):
    pass



# -*- coding: utf-8 -*-

"""
    sijax.helper
    ~~~~~~~~~~~~

    Provides various helper functions and objects that other modules use.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from ..exception import SijaxError


# try to load the best simplejson implementation available.  If JSON
# is not installed, we add a failing class.
json = None
try:
    import simplejson as json
except ImportError:
    try:
        import json
    except ImportError:
        try:
            # Google Appengine offers simplejson via django
            from django.utils import simplejson as json
        except ImportError:
            pass

if json is None:
    raise SijaxError("You need a JSON library to use Sijax!")




# -*- coding: utf-8 -*

"""
    sijax.Sijax
    ~~~~~~~~~~~

    Implements the main Sijax class, an instance of which is used to
    register callbacks, invoke callbacks, setup events handlers, etc.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from .helper import json

from .response.BaseResponse import BaseResponse
from .exception import SijaxError

class Sijax(object):
    
    PARAM_REQUEST = 'sijax_rq'
    PARAM_ARGS = 'sijax_args'
    
    #: Event called immediately before calling the response function
    EVENT_BEFORE_PROCESSING = 'before_processing'

    #: Event called immediately after calling the response function
    EVENT_AFTER_PROCESSING = 'after_processing'

    #: Event called when the function to be called is unknown
    EVENT_INVALID_REQUEST = 'invalid_request'
    
    #: Event called when the function was called in a wrong way (bad args count)
    EVENT_INVALID_CALL = 'invalid_call'

    #: An option when registering callbacks that stores the callback function
    PARAM_CALLBACK = 'callback'

    #: An option when regsitering callbacks that allows the response class to be
    #: changed from :class:`response.BaseResponse` to something else.
    PARAM_RESPONSE_CLASS = 'response_class'

    #: An option when registering callbacks that makes certain extra arguments
    #: be passed automatically to the response function after the obj_response argument
    #: and before the other arguments
    PARAM_ARGS_EXTRA = 'args_extra'
    
    def __init__(self):
        cls = self.__class__

        #: Would contain the request data (usually POST)
        self._data = {}
        
        #: Would contain the callbacks (name => dictionary of params)
        self._callbacks = {}

        #: Stores the various event callbacks
        self._events = {}

        #: The URI where the client should send ajax requests to
        self._request_uri = None

        #: The URI to json2.js or similar to provide JSON support
        #: for browsers that don't have it.. Sijax would load this on demand
        self._json_uri = None
        
        #: Stores a cached version of the arguments to pass to the requested function
        self._request_args = None

        def invalid_request(obj_response, func_name):
            """Handler to be called when we try calling an unknown function."""

            msg = "The action you performed is currently unavailable! (Sijax error)"
            obj_response.alert(msg)

        def invalid_call(obj_response, callback):
            """Handler to be called when we try to call a function in a wrong way."""

            msg = "You tried to perform an action in a wrong way! (Sijax error)"
            obj_response.alert(msg)

        # Register the before/after processing events to not do anything by default
        self.register_event(cls.EVENT_BEFORE_PROCESSING, lambda r: r)
        self.register_event(cls.EVENT_AFTER_PROCESSING, lambda r: r)

        # Register the "error" events to show some alerts by default
        self.register_event(cls.EVENT_INVALID_REQUEST, invalid_request)
        self.register_event(cls.EVENT_INVALID_CALL, invalid_call)
    
    def set_data(self, data):
        """Sets the incoming data dictionary (usually POST).

        Sijax needs this data to determine if the current request is a Sijax request,
        and which callback should be called with which arguments.
        """
        self._data = data
        self._request_args = None # cache invalidation
        return self

    def get_data(self):
        """Returns the incoming data array that the current instance uses."""
        return self._data
    
    def register_callback(self, public_name, callback, response_class=None, args_extra=None):
        """Registers the specified callback function with the given name.

        This "exposes" the callback to the browser by the given public name.

        The optional response class parameter could be used to substitute the regular
        :class:`response.BaseResponse` class used by default. An instance of response_class
        is passed automatically as a first parameter to your response function.

        The optional args_extra parameter allows you (or rather the framework that
        you're using) to pass a list of extra arguments to your response function, immediately
        after the first obj_response argument and before any other user-defined arguments.
        If you pass args_extra=["arg1", "arg2"] when registering a response function,
        the function's signature should look like this:
            def my_func(obj_response, arg1, arg2, {call params here})
        
        :param public_name: the name this function will be exposed to in the browser
        :param callback: the actual function to call
        :param response_class: the obj_response class to use instead of BaseResponse -
                               it could also be a callable that returns a response object
        :param args_extra: an optional list of additional arguments to pass after
                           obj_response and before form_values
        """

        if response_class is None:
            response_class = BaseResponse
        if not hasattr(response_class, '__call__'):
            raise SijaxError("response_class needs to be a callable!")

        if args_extra is None:
            args_extra = []
        else:
            if isinstance(args_extra, tuple):
                args_extra = list(args_extra)
            elif not isinstance(args_extra, list):
                raise SijaxError("args_extra could be a tuple or a list - nothing else!")

        params = {}
        params[self.__class__.PARAM_CALLBACK] = callback
        params[self.__class__.PARAM_RESPONSE_CLASS] = response_class
        params[self.__class__.PARAM_ARGS_EXTRA] = args_extra

        self._callbacks[public_name] = params
        return self
    
    def register_object(self, obj, **options):
        """Registers all "public" callable attributes of the given object.
        
        The object could be anything (module, class, class instance, etc.)

        :param obj: the object whose callable attributes to register
        :param options: the options to be sent to :meth:`register_callback`
        """
        for attr_name in dir(obj):
            if attr_name.startswith('_'):
                continue
            
            attribute = getattr(obj, attr_name)
            if hasattr(attribute, '__call__'):
                self.register_callback(attr_name, attribute, **options)
    
    def is_sijax_request(self):
        """Tells whether this page request looks like a valid request to Sijax.

        Even if this is a sijax request, this doesn't mean it's absolutely valid.
        If the request is determined to be invalid when it gets called later,
        the appropriate events will be fired (EVENT_INVALID_REQUEST/EVENT_INVALID_CALL).
        """
        for k in (self.__class__.PARAM_REQUEST, self.__class__.PARAM_ARGS):
            if k not in self._data:
                return False

        return True
    
    def get_request_function(self):
        """Returns the name of the requested function or None for non-sijax requests."""
        request = self.__class__.PARAM_REQUEST
        return str(self._data[request]) if request in self._data else None
    
    def get_request_args(self):
        """Returns the original arguments list to be passed to the requested function.

        It should be noted that this is not necessarily the arguments list that the
        function will receive. Custom Response objects are allowed to override them.
        Certain plugins make use of this functionality.
        """
        key_args = self.__class__.PARAM_ARGS
        if self._request_args is None:
            # no precached version.. cache it now
            self._request_args = []
            if key_args in self._data:
                try:
                    args = json.loads(self._data[key_args])
                    if isinstance(args, list):
                        self._request_args = args
                except:
                    pass
                
        return self._request_args
    
    def process_request(self):
        """Executes the Sijax request and returns the response.

        If you try calling this for non Sijax requests it would throw a SijaxError.

        If the function that was called from the browser is not registered,
        the EVENT_INVALID_REQUEST handler will be called instead.

        Refer to execute_callback() to see how the main handler is called and
        what the response is.
        """

        if not self.is_sijax_request():
            raise SijaxError("You should not call this for non sijax requests!")
        
        function_name = self.get_request_function()
        if function_name in self._callbacks:
            options = self._callbacks[function_name]
            args = self.get_request_args()
        else:
            # Function not registered.. Let's call the invalid request handler
            # passing to it the function name that should've been called
            args = [function_name]
            callback = self._events[self.__class__.EVENT_INVALID_REQUEST]
            options = {self.__class__.PARAM_CALLBACK: callback}
        
        return self.execute_callback(args, **options)
    
    def execute_callback(self, args, callback, **params):
        """Executes the given callback function and returns a response.

        Before executing the given callback, the EVENT_BEFORE_PROCESSING event
        callback is fired.
        After executing it, the EVENT_AFTER_PROCESSING event callback is fired.

        The response could either be a string (for regular functions) or a
        generator (for streaming functions like Comet or Upload).
        
        :param args: the arguments list to pass to the callback
        :param callback: the callback function to execute
        :param options: more options - look at :meth:`register_callback` to see what else is available
        :return: string for regular callbacks or generator for streaming callbacks
        """
        cls = self.__class__

        # Another response class could be used to extend behavior
        response_class = BaseResponse
        if cls.PARAM_RESPONSE_CLASS in params:
            response_class = params[cls.PARAM_RESPONSE_CLASS]

        args_extra = []
        if cls.PARAM_ARGS_EXTRA in params:
            args_extra = params[cls.PARAM_ARGS_EXTRA]

        if not hasattr(callback, '__call__'):
            raise SijaxError("Provided callback is not callable!")

        # Pass the original call args to the response and get them back later
        # This allows the response class (potentially custom) to override the args
        # Note that we're not passing args_extra to it, as we don't want responses
        # to know anything about that yet
        obj_response = response_class(self, args)
        call_args = args_extra + obj_response.request_args
        
        event_before_processing = cls.EVENT_BEFORE_PROCESSING
        event_after_processing = cls.EVENT_AFTER_PROCESSING

        call_chain = [
            (self._events[event_before_processing], []),
            (callback, call_args),
            (self._events[event_after_processing], [])
        ]
        return obj_response._process_call_chain(call_chain)

    def register_event(self, event_name, callback):
        """Register a callback function to be called when the event occurs.

        Only one callback can be executed per event.

        The provided EVENT_* constants should be used for handling system events.
        Additionally, you can use any string to define your own events and callbacks.

        If more are needed, they may be chained manually, although it's not recommended.
        """
        self._events[event_name] = callback
        return self

    def has_event(self, event_name):
        """Tells whether we've got a registered event by that name."""
        return event_name in self._events

    def get_event(self, event_name):
        """Returns the event handler callback for the specified event."""
        return self._events[event_name] if event_name in self._events else None

    def set_request_uri(self, uri):
        """Specifies the URI where ajax requests will be sent to by the client."""
        self._request_uri = uri
        return self
    
    def set_json_uri(self, uri):
        """
        Sets the URI to an external JSON library,
        for browsers that do not support native JSON (such as IE <= 7).

        If this is not specified, Sijax will not work at all in such browsers.

        The script will only be loaded if a browser without JSON support is detected.
        """
        self._json_uri = uri
        return self
    
    def get_js(self):
        """Returns the javascript needed to prepare the Sijax environment."""

        if self._request_uri is None:
            raise SijaxError("Trying to get the sijax js, but no request_uri is set!")
        
        js = "Sijax.setRequestUri(%s);" % json.dumps(self._request_uri)

        if self._json_uri is not None:
            js += "Sijax.setJsonUri(%s);" % json.dumps(self._json_uri) 

        return js




# -*- coding: utf-8 -*-

"""
    sijax.plugin.upload
    ~~~~~~~~~~~~~~~~~~~

    Provides helpers to register Upload functions.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from ...helper import json
from ...exception import SijaxError
from .UploadResponse import UploadResponse


# Parameters with these names are passed to the client side (JS)
# Modifying this would mean modifying the javascript files
PARAM_FORM_ID = "formId"
PARAM_CALLBACK = "callback"


def _prepare_options(sijax_instance, options):
    param_response_class = sijax_instance.__class__.PARAM_RESPONSE_CLASS
    if param_response_class not in options:
        options[param_response_class] = UploadResponse

    return options


def func_name_by_form_id(form_id):
    return "%s_upload" % form_id


def register_upload_callback(sijax_instance, form_id, callback, **options):
    """Helps you easily register Upload functions with Sijax.
    
    We suggest that the `args_extra` options is passed here, so that the
    response function would receive the files object or anything else it might be
    needing to manipulate file uploads.
    By using `args_extra` you could change the response function's default header from:
        def upload_handler(obj_response, form_values)
    to:
        def upload_handler(obj_response, files, form_values)
    where the args_extra=[files] option was passed during registering.

    It's very important to keep in mind that unlike the regular
    :meth:`Sijax.register_callback` function, this one returns some javascript code
    as a response. You need to execute that code on the page that contains the
    form you want to transform to Sijax upload.

    :param sijax_instance: the :class:`Sijax` instance to register callbacks with
    :param form_id: the id of the form as it appears in the DOM
    :param callback: the function that would get called to process the upload request
    :param options: options to pass to :meth:`Sijax.register_callback`
    :return: string - javascript code you need to put on your page,
                      to make your form use Sijax when submitted
    """

    public_name = func_name_by_form_id(form_id)

    options = _prepare_options(sijax_instance, options)
    sijax_instance.register_callback(public_name, callback, **options)

    js_params = json.dumps({PARAM_FORM_ID: form_id, PARAM_CALLBACK: public_name})
    return "jQuery(function() { sjxUpload.registerForm(%s); });" % js_params




# -*- coding: utf-8 -*-

"""
    sijax.plugin.upload.UploadResponse
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Provides the UploadResponse class, an instance of which
    is passed as the first argument to all Upload functions.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from ...response.StreamingIframeResponse import StreamingIframeResponse


class UploadResponse(StreamingIframeResponse):

    def __init__(self, *args, **kwargs):
        StreamingIframeResponse.__init__(self, *args, **kwargs)

        # The form_id that we're dealing with is the only expected argument.
        req_args = self._request_args
        req_args_len = len(req_args)
        if req_args_len != 1:
            raise RuntimeError("We only expect 1 request argument, %d given." % req_args_len)
        self._form_id = req_args[0]

        # Let's override the request arguments now
        # The upload handler function is to expect only one argument
        # which is a dictionary of form values
        # To get the form values from all the POST data,
        # we simply need to get rid of some internal/system params

        # We're creating a new dict, because the POST data dictionary
        # may be immutable and/or we don't want to change it anyways.
        # We want to work on a copy of it.
        form_values = dict(self._sijax.get_data())
        del form_values[self._sijax.__class__.PARAM_REQUEST]
        del form_values[self._sijax.__class__.PARAM_ARGS]
        self._request_args = [form_values]

    @property
    def form_id(self):
        return self._form_id

    def reset_form(self):
        return self.call("sjxUpload.resetForm", self.form_id)



# -*- coding: utf-8 -*-

"""
    sijax.response.BaseResponse
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Provides the BaseResponse class, which is the default class
    used to create response objects from for regular Sijax functions.
    The response class is the way for Sijax functions (handlers) to
    pass information back to the browser. They do this by calling
    various methods, which queue commands until they're sent to the browser.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from ..helper import json


class BaseResponse(object):
    
    COMMAND_ALERT = 'alert'
    COMMAND_HTML = 'html'
    COMMAND_SCRIPT = 'script'
    COMMAND_ATTR = 'attr'
    COMMAND_CSS = 'css'
    COMMAND_REMOVE = 'remove'
    COMMAND_CALL = 'call'
    
    def __init__(self, sijax_instance, request_args):
        """Constructs a new empty Sijax Response object.

        The arguments that the sijax function is invoked with are passed,
        so that we can override them if we want.
        We don't need to do that for regular sijax responses,
        but certain plugins may need to do that.

        :param sijax_instance: the Sijax instance object that created this response
        :param request_args: the request arguments that the sijax function
                             was invoked with
        """
        self._commands = []
        self._sijax = sijax_instance
        self._request_args = request_args
    
    @property
    def request_args(self):
        """Returns the arguments array to pass to callbacks.

        This class doesn't manipulate them, so it returns the original arguments,
        but other classes can override them however they need.
        """
        return self._request_args
        
    def add_command(self, cmd_type, params = None):
        """Adds a raw command to the buffer to send to the client.

        You generally don't need to use this at all,
        unless you really know what you're doing.
        """
        if params is None:
            params = {}
        params['type'] = cmd_type
        
        self._commands.append(params)
        return self
    
    def clear_commands(self):
        """Clears the commands buffer."""
        self._commands = []
        return self

    def alert(self, message):
        """Sends a window.alert command to the browser."""
        params = {self.__class__.COMMAND_ALERT: message}
        return self.add_command(self.__class__.COMMAND_ALERT, params)
    
    def _html(self, selector, html, set_type):
        params = {"selector": selector, "html": html, "setType": set_type}
        return self.add_command(self.__class__.COMMAND_HTML, params)
   
    def html(self, selector, html):
        """
        Adds the given html to the element specified by the selector,
        replacing any html content inside it.

        Scripts inside the html block are also executed.

        :param selector: the jQuery selector for which we'll replace the html
        :param html: the html text
        """
        return self._html(selector, html, "replace")

    def html_append(self, selector, html):
        """Appends the given html to the element specified by the selector.

        Scripts inside the html block are also executed.

        :param selector: the jQuery selector to append html to
        :param html: the html text
        """
        return self._html(selector, html, "append")
   
    def html_prepend(self, selector, html):
        """Prepends the given html to the element specified by the selector.

        Scripts inside the html block are also executed.

        :param selector: the jQuery selector to prepend html to
        :param html: the html text
        """
        return self._html(selector, html, "prepend")
    
    def script(self, js):
        """Executes the given javascript code.

        Note that the given javascript code is eval-ed inside a
        Sijax helper function, so it's not touching the global namespace,
        unless you explicitly do it.
        """
        params = {self.__class__.COMMAND_SCRIPT: js}
        return self.add_command(self.__class__.COMMAND_SCRIPT, params)

    def css(self, selector, property_name, value):
        """Assigns a new style property value to all elements matching the selector.

        Finds an element by the specified selector and changes
        the specified css (style) property to the given value.
        Same as jQuery's $(selector).css('property', 'value');

        Note that this can only change a single property at once.
        It doesn't support jQuery's mass change which uses a
        key/value map of properties/values to change.

        :param selector: the jQuery selector to invoke css() on
        :param property_name: the name of the style property
        :param value: the new value to assign to the property
        """
        params = {"selector": selector, "key": property_name, "value": value}
        return self.add_command(self.__class__.COMMAND_CSS, params)

    def _attr(self, selector, property_name, value, set_type):
        params = {
            "selector": selector, "key": property_name,
            "value": value, "setType": set_type
        }
        return self.add_command(self.__class__.COMMAND_ATTR, params)

    def attr(self, selector, property_name, value):
        """Assigns an attribute value to all elements matching the selector.

        Finds an element by the specified selector and changes
        the specified property to the given value.
        Same as jQuery's $(selector).attr('property', 'value');

        :param selector: the jQuery selector to invoke attr() on
        :param property_name: the name of the property
        :param value: the new value to assign to the property
        """
        return self._attr(selector, property_name, value, "replace")
    
    def attr_append(self, selector, property_name, value):
        """Same as attr(), but appends instead of assigning a new value."""
        return self._attr(selector, property_name, value, "append")

    def attr_prepend(self, selector, property_name, value):
        """Same as attr(), but prepends instead of assigning a new value."""
        return self._attr(selector, property_name, value, "prepend")
    
    def remove(self, selector):
        """Removes all elements from the DOM that match the selector."""
        params = {self.__class__.COMMAND_REMOVE: selector}
        return self.add_command(self.__class__.COMMAND_REMOVE, params)

    
    def redirect(self, uri):
        """Redirects the browser to the given URI."""
        return self.script("window.location = %s;" % json.dumps(uri))
    
    def call(self, js_func_name, func_params=None):
        """Calls the given javascript function with the given arguments list.

        :param js_func_name: the name of the javascript function to call
        :param func_params: a list of arguments to call the function with
        """
        if func_params is None:
            func_params = []
        params = {self.__class__.COMMAND_CALL: js_func_name, "params": func_params}
        return self.add_command(self.__class__.COMMAND_CALL, params)
    
    def get_json(self):
        """Returns a JSON representation of the commands buffer list.

        The client side code will loop over the list and execute all the
        commands in order.
        """
        return json.dumps(self._commands)
    
    def _process_callback(self, callback, args):
        """Processes a single callback.

        For normal responses this only means calling the callback,
        because we'll flush the commands to the client once (in the end).

        For other responses (like Comet) though we can flush implicitly after
        every callback.
        """
        try:
            callback(self, *args)
        except TypeError:
            # we should re-raise the exception if it's coming
            # from within the function, meaning calling works.. @todo
            event_invalid_call = self._sijax.__class__.EVENT_INVALID_CALL
            self._sijax.get_event(event_invalid_call)(self, callback)

    def _process_call_chain(self, call_chain):
        """Executes all the callbacks in the chain for this response object.

        Each callback in the chain adds commands to the buffer.
        When all the callbacks have been executed, the buffer contains a list of commands
        that we need to pass to the browser (in order).

        The result of this is a JSON string to pass to the browser.

        :param call_chain: a list of two-tuples (callback, args list) to call
        """
        for callback, args in call_chain:
            self._process_callback(callback, args)
        return self.get_json()




# -*- coding: utf-8 -*-

"""
    sijax.plugin.comet
    ~~~~~~~~~~~~~~~~~~

    Provides helpers to register Comet functions.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from .CometResponse import CometResponse


def _prepare_options(sijax_instance, options):
    param_response_class = sijax_instance.__class__.PARAM_RESPONSE_CLASS
    if param_response_class not in options:
        options[param_response_class] = CometResponse

    return options


def register_comet_callback(sijax_instance, public_name, callback, **options):
    """Helps you easily register Comet functions with Sijax.
    
    :param sijax_instance: the :class:`Sijax` instance to register callbacks with
    :param public_name: the name of the function that the client will use to call it
    :param callback: the actual function that would get called to process the request
    :param options: options to pass to :meth:`Sijax.register_callback`
    """
    options = _prepare_options(sijax_instance, options)
    sijax_instance.register_callback(public_name, callback, **options)


def register_comet_object(sijax_instance, obj, **options):
    """Helps you easily register all "public" callable attributes of an object.

    :param sijax_instance: the :class:`Sijax` instance to register callbacks with
    :param obj: the object whose methods to register with Sijax
    :param options: options to pass to :meth:`Sijax.register_callback`
    """
    options = _prepare_options(sijax_instance, options)
    sijax_instance.register_object(obj, **options)
        



"""
Sijax
=====

Sijax stands for "Simple ajax" and provides just that.
It's a simple Python/jQuery library providing easy ajax integration for python web apps.

The main idea is to use javascript code that calls server-side callbacks, which generate a response (manipulating the DOM, etc) and pass it back to the client.
This way, you don't need to manually dispatch ajax requests to certain URIs and go over each XML/JSON response manually.

Here's a tiny snippet of code to show what it's capable of doing::

    # Function definition in Python
    def say_hello_handler(obj_response, hello_from, hello_to):
        obj_response.alert("Hello from %s to %s" % (hello_from, hello_to))
        obj_response.alert("Redirecting you..")
        obj_response.redirect("https://github.com/spantaleev/sijax-python")

    # Expose the above function publicly by the name of "say_hello"
    sijax_instance.register_callback("say_hello", say_hello_handler)

    //The above function can be called from javascript using
    Sijax.request('say_hello', ['John', 'Greg']);
"""

from setuptools import setup, find_packages

import sijax


def run_tests():
    import os, sys
    sys.path.append(os.path.join(os.path.dirname(__file__), 'tests'))
    from sijax_tests import suite
    return suite()


setup(
    name = "Sijax",
    packages = find_packages(),
    include_package_data = True,
    version = sijax.__version__,
    description = "An easy to use AJAX library based on jQuery.ajax",
    long_description = __doc__,
    author = "Slavi Pantaleev",
    author_email = "s.pantaleev@gmail.com",
    url = "https://github.com/spantaleev/sijax-python",
    keywords = ["ajax", "jQuery"],
    platforms = "any",
    license = "BSD",
    zip_safe = False,
    classifiers = [
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    test_suite='__main__.run_tests'
)



# -*- coding: utf-8 -*-

"""
    sijax.response.StreamingIframeResponse
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Provides a response class to be used when Sijax functions
    are invoked via an iframe, instead of XHR.
    The main difference from the BaseResponse class is that this one
    can do streaming, meaning it can flush the commands queue back to
    the browser at any time.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from .BaseResponse import BaseResponse
from types import GeneratorType


class StreamingIframeResponse(BaseResponse):
    """A response class used with iframe-calls and supporting streaming.

    This is used by plugins that use an iframe to perform the Sijax call,
    instead of doing a regular ajax request.
    This has the benefit of allowing us to do streaming, which means that
    response functions can flush the commands buffer whenever they want
    without exiting.
    """

    def __init__(self, *args, **kwargs):
        BaseResponse.__init__(self, *args, **kwargs)
        self._is_first_flush = True
    
    def flush(self):
        """Generates command output to flush to the browser.
        
        The output is not JSON, because it's evaluated in an
        iframe. We're generating some html markup with script tags
        to pass our commands JSON to the parent, which will then execute it.

        We're also pushing some garbage content on the first flush,
        to get around a buffering "issue" certain browsers employ.
        Such browsers include IE and Google Chrome.
        They're generally buffering the first ~1500 bytes of data,
        before they start interpretting it.

        This should be used in streaming (Comet) functions when data
        is to be flushed to the browser.
        It should be used like this:
            obj_response.alert("Message")
            yield obj_response
            ... some other code here..
        which sends the alert() to the browser before executing the other code.

        Your don't need to explicitly flush at the end of a streaming function.
        What remains unflushed when the function exits will eventually get flushed.
        """
        output = """
        <script type="text/javascript">
            window.parent.Sijax.processCommands(%s);
        </script>
        """ % self.get_json()

        self.clear_commands()

        if self._is_first_flush:
            # Push some more data initially, because certain
            # browsers buffer the first X bytes of output
            self._is_first_flush = False
            output = "%s%s%s" % (
                "\n<script type='text/javascript'></script>\n\n",
                "\n" * 2000,
                output
            )

        return output
    
    def _process_callback(self, callback, args):
        """Processes a callback to a normal or a streaming function.

        In constrast with SijaxResponse._process_callback, which only
        processes normal requests properly, this can process normal
        and streaming functions.

        Normal functions are the typical Sijax response functions,
        which don't flush content to the browser, but only push commands
        to the buffer list. Those commands are to be flushed by Sijax
        when the response function exits.

        Streaming functions are the typical Comet response functions,
        which are generators (they use yield to flush content).

        Basically this function can be seens as a converter from
        either a generator (streaming function) or a string (normal function)
        to a generator.
        """
        try:
            response = callback(self, *args)
        except TypeError:
            # we should re-raise the exception if it's coming
            # from within the function, meaning calling works.. @todo
            failed_callback = callback
            event_invalid_call = self._sijax.__class__.EVENT_INVALID_CALL
            callback = self._sijax.get_event(event_invalid_call)
            response = callback(self, failed_callback)


        if isinstance(response, GeneratorType):
            # Real streaming function using a generator to flush
            while True:
                # we don't really care what it yields..
                response.next()
                if len(self._commands) != 0:
                    yield self.flush()
        else:
            # Normal (non-streaming) function
            # Let's flush implicitly for such functions
            if len(self._commands) != 0:
                yield self.flush()

    def _process_call_chain(self, call_chain):
        """Executes all the callbacks in the chain for streaming response objects.

        The difference from SijaxResponse._process_call_chain is that this
        returns a generator instead of a string. This allows response functions
        to flush the commands buffer whenever they need, instead of all at once
        in the end.

        :param call_chain: a list of two-tuples (callback, args list) to call
        """
        for callback, args in call_chain:
            generator = self._process_callback(callback, args)
            for string in generator:
                yield str(string)

